#!/usr/bin/env python3
"""
    TRAINER SCRIPT - contains custom code for training your model
"""
import os
import sys
import json
import argparse
import logging

# Bring in Sagemaker Container Environment handling
from mldock.platform_helpers.aws.sagemaker import environment
env = environment.Environment()

def str2bool(arg_value: str) -> bool:
    if isinstance(arg_value, bool):
        return arg_value
    if arg_value.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif arg_value.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

def str2dict(arg_value: str) -> dict:
    d_argvalue = json.loads(arg_value)
    if isinstance(d_argvalue, dict):
        return d_argvalue
    else:
        raise argparse.ArgumentTypeError('Json Dictionary expected.')

def run_training():
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger('sagemaker')
    logger.setLevel(logging.INFO)
    PARSER= argparse.ArgumentParser()
    ## User set params which link to Sagemaker params
    PARSER.add_argument('--training', type=str,
                        default=env.channel_input_dirs["training"],
                        help='Expects a s3 path to directory containing a sparse .npz file containing item user matrix')
    PARSER.add_argument('--factors', default=env.hyperparameters.get('factors', 20),
                        type=int,
                        help='Number of factorization factors to compute during model fit.')
    PARSER.add_argument('--regularization', default=env.hyperparameters.get('regularization', 0.3),
                        type=int,
                        help='Coefficient of regularization to apply during model fit.')
    PARSER.add_argument('--iterations', default=env.hyperparameters.get('iterations', 10),
                        type=int,
                        help='Number iterations of the dataset should be run during model fit.')
    PARSER.add_argument('--run_eval', default=env.hyperparameters.get('run_eval', True),
                        type=str2bool,
                        help='Whether to run metrics evaluation prior to training model on full data.')
    PARSER.add_argument('--refit', default=env.hyperparameters.get('refit', True),
                        type=str2bool,
                        help='Whether or not to save model.')

    # Get args
    ARGS, _ = PARSER.parse_known_args()

    if ARGS.run_eval:
        logger.info('Running Evaluation')
        # TODO Add your Evaluation Fitting here
        
        logger.info("Your data is here = {}".format(ARGS.training))

        logger.info("Some hyperparameters = factors={}, regularization={}, iterations={}".format(ARGS.factors, ARGS.regularization, ARGS.iterations))


    # save model to s3
    if ARGS.refit:
        logger.info('Refitting Final Model')
        # TODO Add model.fit(X) here again


        # save model to s3
        logger.info("Model saved in {}. This is written back to s3 at job completion.".format('/opt/ml/model'))
        model_dir = os.path.join('/opt/ml/model','model.pkl')
        # TODO add model saving workflow here
        pass
    else:
        logger.info("Skipping model save step. ")

if __name__ == '__main__':

    run_training()



import os
import sys
import traceback

from mldock.platform_helpers.environment import Environment
import logging

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger('mldock')

environment = Environment(base_dir='/opt/ml')
MODEL_DIR = environment.model_dir

def extract_stack_trace() -> str:
    """Extracts full stacktrace from thown exception
    Returns:
        str: traceback of error thrown
    """
    exc_type, exc_value, exc_traceback = sys.exc_info()
    stack_trace = traceback.format_exception(
        exc_type,
        exc_value,
        exc_traceback
    )
    return repr(stack_trace)

class ModelService(object):
    model = None


    @classmethod
    def get_model(cls):
        """Get the model object for this instance, loading it if it's not already loaded."""
        if cls.model == None:
            logger.info("Loading Model Artifacts")
            # TODO place your model load code here
            # get your model from MODEL_DIR
            cls.model = None
        return cls.model

    @classmethod
    def predict(cls, input):
        """For the input, do the predictions and return them.

        Args:
            input (a pandas dataframe): The data on which to do the predictions. There will be
                one prediction per row in the dataframe"""
        model = cls.get_model()
        
        try:
            logger.info("Running Prediction")
            # TODO run model prediction has required

            # TODO Add any post processing

            return None
        except Exception as exception:
            # get stack trace as exception
            stack_trace = extract_stack_trace()
            reformatted_log_msg = (
                    'Server Error: {ex}'.format(ex=stack_trace)
            )
            return reformatted_log_msg


def predict(json_input):
    """
    Prediction given the request input
    :param json_input: [dict], request input
    :return: [dict], prediction
    """

    # TODO Transform json_input and assign the transformed value to model_input
    model_input = None
    
    result = ModelService.predict(model_input)
    logger.debug(result)

    return result



import os
import sys
import logging
import click

from mldock.platform_helpers.docker.auth import login_and_authenticate
from mldock.api.local import \
    docker_build
from mldock.api.cloud import \
    push_image_to_repository
from mldock.config.config_manager import \
    MLDockConfigManager

click.disable_unicode_literals_warning = True
logger=logging.getLogger('mldock')

@click.group()
def cloud():
    """
    Commands for local operations: train and deploy
    """
    pass

@click.command()
@click.option(
    '--dir',
    help='Set the working directory for your mldock container.',
    required=True,
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None
    )
)
@click.option('--no-cache', help='builds container from scratch', is_flag=True)
@click.option(
    '--build',
    help='Set the working directory for your mldock container.',
    is_flag=True
)
@click.option(
    '--provider',
    help='Set the cloud provider',
    required=True,
    type=click.Choice(['ecr'],
    case_sensitive=False
    )
)
@click.pass_obj
def push(obj, dir, build, provider, no_cache):
    """
    Command to push docker container image to ECR using your default AWS sdk credentials.

    note: Override default AWS credentials by providing AWS_DEFAULT_PROFILE environment variable.
    """
    helper_library_path = obj['helper_library_path']
    try:
        mldock_manager = MLDockConfigManager(
            filepath=os.path.join(dir, ".mldock.json")
        )
        # get mldock_module_dir name
        mldock_config = mldock_manager.get_config()
        image_name = mldock_config.get("image_name", None)
        container_dir = mldock_config.get("container_dir", None)
        module_path = os.path.join(
            dir,
            mldock_config.get("mldock_module_dir", "src"),
        )
        dockerfile_path = os.path.join(
            dir,
            mldock_config.get("mldock_module_dir", "src"),
            container_dir
        )
        requirements_file_path = os.path.join(
            dir,
            mldock_config.get("requirements.txt", "requirements.txt")
        )
        if image_name is None:
            raise Exception("\nimage_name cannot be None")
        elif image_name.endswith(":latest"):
            raise Exception("\nImage version is not supported at this point. Please remove :latest versioning")
        else:

            # login and authenticate
            client, metadata = login_and_authenticate(provider=provider, region='eu-west-1')
            image_repository = "{}/{}".format(metadata['repository'], image_name)
            logger.info("\nLogin Complete! ヽ(´▽`)/\n")

            # build image for cloud repository
            docker_build(
                image_name=image_repository,
                dockerfile_path=dockerfile_path,
                module_path=module_path,
                target_dir_name=mldock_config.get("mldock_module_dir", "src"),
                requirements_file_path=requirements_file_path,
                no_cache=no_cache
            )

            # Push image to cloud repository
            push_image_to_repository(
                image_repository=image_repository,
                auth_config = {'username': metadata['username'], 'password': metadata['password']},
                tag='latest'
            )
            logger.info("\nPush Complete! ヽ(´▽`)/\n")

    except Exception as e:
        logger.error("{}".format(e))
        sys.exit(-1)


cloud.add_command(push)



import os
import sys
import json
import logging
from pathlib import Path
import docker
import traceback
import logging
import signal

from distutils.dir_util import copy_tree, remove_tree
from distutils.dir_util import mkpath
from distutils.file_util import write_file
from future.moves import subprocess

logger=logging.getLogger('mldock')


def train_model(working_dir, docker_tag, image_name, entrypoint, cmd, env=None):
    """
    Trains ML model(s) locally
    :param working_dir: [str], source root directory
    :param docker_tag: [str], the Docker tag for the image
    :param image_name: [str], The name of the Docker image
    """
    process_env = None
    if env is not None:
        process_env = os.environ.copy()
        process_env.update(env)

    client = docker.from_env()

    try:
        container = client.containers.run(
            image="{image}:{tag}".format(image=image_name, tag=docker_tag),
            entrypoint=entrypoint,
            command=cmd,
            environment=process_env,
            remove=True,
            tty=True,
            volumes={
                '{test_path}/config'.format(test_path=os.path.abspath(working_dir)): {'bind': '/opt/ml/input/config', 'mode': 'rw'},
                '{test_path}/data'.format(test_path=os.path.abspath(working_dir)): {'bind': '/opt/ml/input/data', 'mode': 'rw'},
                '{test_path}/model'.format(test_path=os.path.abspath(working_dir)): {'bind': '/opt/ml/model', 'mode': 'rw'},
                '{test_path}/output'.format(test_path=os.path.abspath(working_dir)): {'bind': '/opt/ml/output', 'mode': 'rw'}
            },
            auto_remove=True,
            detach=True,
            stream=True
        )
        logs = container.logs(follow=True).decode('utf-8')

        logger.info(logs)
    except (KeyboardInterrupt, SystemExit) as exception:
        logger.error(exception)
        container.kill()
    except (docker.errors.APIError, docker.errors.ContainerError, docker.errors.ImageNotFound) as exception:
        logger.error(exception)
    except Exception as exception:
        logger.error(exception)

def deploy_model(working_dir, docker_tag, image_name, entrypoint, cmd, port=8080, env={}):
    """
    Deploys ML models(s) locally
    :param working_dir: [str], source root directory
    :param docker_tag: [str], the Docker tag for the image
    :param image_name: [str], The name of the Docker image
    """

    process_env = os.environ.copy()
    if isinstance(env, dict) & len(env) > 0:
        process_env.update(env)
    client = docker.from_env()

    try:
        container = client.containers.run(
            image="{image}:{tag}".format(image=image_name, tag=docker_tag),
            entrypoint=entrypoint,
            command=cmd,
            environment=process_env,
            ports={8080: port},
            remove=True,
            tty=True,
            volumes={
                '{test_path}/config'.format(test_path=os.path.abspath(working_dir)): {'bind': '/opt/ml/input/config', 'mode': 'rw'},
                '{test_path}/data'.format(test_path=os.path.abspath(working_dir)): {'bind': '/opt/ml/input/data', 'mode': 'rw'},
                '{test_path}/model'.format(test_path=os.path.abspath(working_dir)): {'bind': '/opt/ml/model', 'mode': 'rw'},
                '{test_path}/output'.format(test_path=os.path.abspath(working_dir)): {'bind': '/opt/ml/output', 'mode': 'rw'}
            },
            auto_remove=True,
            detach=True,
            stream=True
        )
        logs = container.logs(follow=True).decode('utf-8')

        logger.info(logs)
    except (KeyboardInterrupt, SystemExit) as exception:
        logger.error(exception)
        container.kill()
    except (docker.errors.APIError, docker.errors.ContainerError, docker.errors.ImageNotFound) as exception:
        logger.info(exception)
        logger.error(exception)
    except Exception as exception:
        logger.error(exception)

def pretty_build_logs(line: dict):

    if line is None:
        return None
    error = line.get('error', None)
    errorDetail = line.get('errorDetail', None)
    if error is not None:
        logger.error('{}\n{}'.format(error, errorDetail))
    
    stream = line.get('stream','')

    if ('Step' in stream) & (':' in stream):
        logger.info(stream)
    else:
        logger.debug(stream)

def docker_build(
    image_name: str,
    dockerfile_path: str,
    module_path: str,
    target_dir_name: str,
    requirements_file_path: str,
    no_cache: bool,
    client = None
):
    """Runs the build executable from script path, passing a set of arguments in a command line subprocess.

    Args:
        script_path (str): relative path to script when run on root
        base_path (str):
        image_name (str):
        dockerfile_path (str):
        module_path (str):
        target_dir_name (str):
        requirements_file_path (str):
    """

    # if client is None:
        # client = docker.from_env()
    client = docker.APIClient(base_url=os.environ.get('DOCKER_HOST','tcp://127.0.0.1:1234'))

    try:
        short_log = []

        logs = client.build(
            tag=image_name,
            path='.',
            dockerfile=os.path.join(dockerfile_path, 'Dockerfile'),
            buildargs={
                'module_path': module_path,
                'target_dir_name': target_dir_name,
                'requirements_file_path': requirements_file_path
            },
            quiet=False,
            nocache=no_cache,
            rm=True,
            decode=True
        )

        for line in logs:
            pretty_build_logs(line=line)
    except (KeyboardInterrupt, SystemExit) as exception:
        logger.error(exception)
    except (docker.errors.APIError, docker.errors.ContainerError, docker.errors.ImageNotFound) as exception:
        logger.info(exception)
        logger.error(exception)
    except Exception as exception:
        logger.error(exception)

def _write_file(filepath, parents=True):
    """write a file"""
    if parents == True:
        mkpath(str(Path(filepath).parents[0].absolute()))
    write_file(str(Path(filepath).absolute()), '')

def _rename_file(base_path, current_filename, new_filename):
    """renames filename for a given base_path, saving the file in the same base_path

    Args:
        base_path ([type]): directory path containing file to rename
        current_filename ([type]): current name of the file to rename
        new_filename ([type]): new name for the renamed file
    """
    Path(base_path, current_filename).rename(Path(base_path, new_filename))

def _create_empty_file(base_path, filename):
    """renames filename for a given base_path, saving the file in the same base_path

    Args:
        base_path ([type]): directory path containing file to rename
        current_filename ([type]): current name of the file to rename
        new_filename ([type]): new name for the renamed file
    """
    Path(base_path, filename).touch(exist_ok=True)

def _copy_boilerplate_to_dst(src: str, dst: str, remove_first=False):
    """[summary]

    Args:
        src (str): [description]
        dst (str): [description]
    """
    source_path = str(Path(src).absolute())
    destination_path = str(Path(dst).absolute())
    if remove_first == True:
        try:
            logger.debug("removing first")
            remove_tree(destination_path)
        except FileNotFoundError as exception:
            logger.debug("No file found. Assuming already deleted.")
    response = copy_tree(source_path, destination_path)



"""

"""
"""
    AWS Auth utilities
"""
import base64
import boto3

def get_aws_ecr(region: str):
    # this loads AWS access token and secret from env and returns an ECR client
    ecr_client = boto3.client('ecr', region_name=region)
    # get login token
    token = ecr_client.get_authorization_token()
    username, password = base64.b64decode(token['authorizationData'][0]['authorizationToken']).decode().split(':')
    password=base64.b64decode(token["authorizationData"][0]["authorizationToken"]).decode().split(":")[1]
    registry = token['authorizationData'][0]['proxyEndpoint']
    # return docker credentials
    return username, password, registry



import os
import sys
import io
import csv
import traceback
import logging
import pandas as pd
import numpy as np

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(None)
logger.setLevel(logging.INFO)

MODEL_DIR = '/opt/ml/model'

def extract_stack_trace() -> str:
    """Extracts full stacktrace from thown exception
    Returns:
        str: traceback of error thrown
    """
    exc_type, exc_value, exc_traceback = sys.exc_info()
    stack_trace = traceback.format_exception(
        exc_type,
        exc_value,
        exc_traceback
    )
    return repr(stack_trace)

def array_to_csv(array_like):
    """Convert an array-like object to CSV.
    To understand better what an array-like object is see:
    https://docs.scipy.org/doc/numpy/user/basics.creation.html#converting-python-array-like-objects-to-numpy-arrays
    Args:
        array_like (np.array or Iterable or int or float): array-like object
            to be converted to CSV.
    Returns:
        (str): object serialized to CSV
    """
    stream = io.StringIO()
    
    np.savetxt(stream, array_like, delimiter=",", fmt="'%s'")
    return stream.getvalue()

class ModelService(object):
    model = None

    @classmethod
    def get_model(cls):
        """Get the model object for this instance, loading it if it's not already loaded."""
        if cls.model == None:
            # TODO place your model load code here
            # get your model from MODEL_DIR
            cls.model = None
        return cls.model

    @staticmethod
    def input_load(data):
        """Get the model object for this instance, loading it if it's not already loaded."""
        if isinstance(data, list):
            return pd.DataFrame(data)
        elif isinstance(data, dict):
            return pd.DataFrame([data])
        elif isinstance(data, bytes):
            return pd.read_csv(io.BytesIO(data), sep=",")
        else:
            raise TypeError("Expected payload to have type list or dict")
    
    @staticmethod
    def input_transform(data: pd.DataFrame, cols: list) -> pd.DataFrame:
        """transform the input dataset

        Args:
            data (pd.DataFrame): dataframe of input data

        Returns:
            pd.DataFrame: transformed data
        """
        return data.filter(cols, axis=1)

    @staticmethod
    def output_transform_json(pred):
        """transform numpy array of predictions into json payload"""
        results = pred.tolist()
        return {'results': results}

    @staticmethod
    def output_transform_csv(pred):
        """Transform numpy array of predictions in to csv"""
        predictions = pred.tolist()

        # create results
        results = [('idx','prediction')]
        # append values
        [
            results.append(
                (i,
                pred)
            ) for (i, pred) in enumerate(predictions)
        ]

        return array_to_csv(results)

    @classmethod
    def predict(cls, input, content_type='json'):
        """For the input, do the predictions and return them.

        Args:
            input (a pandas dataframe): The data on which to do the predictions. There will be
                one prediction per row in the dataframe"""
        model = cls.get_model()
        
        try:
            # load input data
            data = cls.input_load(data=input)

            # TODO run input processing as needed
            # keep this light
            # input_transform(data=data, cols=[])

            # TODO run model prediction has required
            # pred = model.predict(data)
            pred = np.array(['pred_1', 'pred_2', 'pred_3'])

            # TODO Add any post processing
            if content_type == 'csv':
                results = cls.output_transform_csv(pred)
            else:
                results = cls.output_transform_json(pred)
            return results

        except Exception as exception:
            # get stack trace as exception
            stack_trace = extract_stack_trace()
            reformatted_log_msg = (
                    'Server Error: {ex}'.format(ex=stack_trace)
            )
            return reformatted_log_msg


def predict(json_input, content_type='json'):
    """
    Prediction given the request input
    :param json_input: [dict], request input
    :return: [dict], prediction
    """

    # TODO Transform json_input and assign the transformed value to model_input
    model_input = json_input
    
    result = ModelService.predict(model_input, content_type=content_type)
    logger.info(result)

    return result



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import os
import sys
import logging
import click

# add mldock package to path
sys.path.insert(1, ".")

from mldock.__version__ import __version__
from mldock.__init__ import MLDOCK_LOGO
# from mldock.command.initialize import init
from mldock.command.container import container
from mldock.command.local import local
from mldock.command.cloud import cloud
from mldock.log import configure_logger

click.disable_unicode_literals_warning = True

CLI_VERSION = 'cli {}'.format(__version__)

@click.group()
@click.option(u"-v", u"--verbose", count=True, help=u"Turn on debug logging")
@click.option(u"-t", u"--docker-tag", default=u"latest", help=u"Specify tag for Docker image")
@click.pass_context
def cli(ctx, verbose, docker_tag):
    """
    CLI tool built to add simple functionality to sagify development.
    """
    _FILE_DIR_PATH = os.path.dirname(os.path.realpath(__file__))
    logger=configure_logger('mldock', verbose=verbose)
    logger.info(MLDOCK_LOGO)
    logger.info('Version: ' + CLI_VERSION)
    ctx.obj = {'verbose': verbose, 'docker_tag': docker_tag, 'helper_library_path':_FILE_DIR_PATH, 'logo':MLDOCK_LOGO}

def add_commands(cli):
    cli.add_command(container)
    cli.add_command(local)
    cli.add_command(cloud)

add_commands(cli)

if __name__ == '__main__':
    cli()



import os
import sys
import json
import logging
import click
from pathlib import Path
from future.moves import subprocess

import mldock.api.predict as predict_request
from mldock.platform_helpers.mldock.utils import \
    collect_mldock_environment_variables
from mldock.config.config_manager import \
    ResourceConfigManager, MLDockConfigManager, PackageConfigManager, \
        HyperparameterConfigManager, InputDataConfigManager
from mldock.api.local import \
    _copy_boilerplate_to_dst, docker_build, \
        _rename_file

from mldock.api.local import \
    train_model, deploy_model

click.disable_unicode_literals_warning = True
logger=logging.getLogger('mldock')

@click.group()
def local():
    """
    Commands for local operations: train and deploy
    """
    pass

def pretty_build_logs(line: dict):

    if line is None:
        return None
    error = line.get('error', None)
    errorDetail = line.get('errorDetail', None)
    if error is not None:
        logger.error('{}\n{}'.format(error, errorDetail))
    
    stream = line.get('stream','')

    if ('Step' in stream) & (':' in stream):
        logger.info("\n"+stream)
    else:
        logger.debug(stream)

@click.command()
@click.option('--dir', help='Set the working directory for your mldock container.', required=True)
@click.option('--no-cache', help='builds container from scratch', is_flag=True)
@click.pass_obj
def build(obj, dir, no_cache):
    """build image using docker

    Args:
        dir (str): directory containing model assets
    """
    helper_library_path = obj['helper_library_path']
    try:
        mldock_manager = MLDockConfigManager(
            filepath=os.path.join(dir, ".mldock.json")
        )
        # get mldock_module_dir name
        mldock_config = mldock_manager.get_config()
        image_name = mldock_config.get("image_name", None)
        container_dir = mldock_config.get("container_dir", None)
        module_path = os.path.join(
            dir,
            mldock_config.get("mldock_module_dir", "src"),
        )
        dockerfile_path = os.path.join(
            dir,
            mldock_config.get("mldock_module_dir", "src"),
            container_dir
        )
        requirements_file_path = os.path.join(
            dir,
            mldock_config.get("requirements.txt", "requirements.txt")
        )
        if image_name is None:
            raise Exception("\nimage_name cannot be None")
        elif image_name.endswith(":latest"):
            raise Exception("\nImage version is not supported at this point. Please remove :latest versioning")
        else:
            logger.info("\nStarting build...\n")
            docker_build(
                image_name=image_name,
                dockerfile_path=dockerfile_path,
                module_path=module_path,
                target_dir_name=mldock_config.get("mldock_module_dir", "src"),
                requirements_file_path=requirements_file_path,
                no_cache=no_cache
            )
            logger.info("\nBuild Complete! ヽ(´▽`)/\n")
    except subprocess.CalledProcessError as e:
        logger.debug(e.output)
        raise
    except Exception as e:
        logger.error("{}".format(e))
        sys.exit(-1)

@click.command()
@click.option('--payload', default=None, help='path to payload file', required=True)
@click.option('--content-type', help='format of payload', type=click.Choice(['json', 'csv'], case_sensitive=False))
@click.option('--host', help='host url at which model is served', type=str, default='http://127.0.0.1:8080/invocations')
def predict( payload, content_type, host):
    """
    Command to run curl predict against an host served on localhost:8080
    """
    try:
        if payload is None:
            logger.info("\nPayload cannot be None. Please provide path to payload file.")
        else:
            if content_type in ['application/json', 'json']:
                pretty_output = predict_request.send_json(filepath=payload, host=host)
            elif content_type in ['text/csv', 'csv']:
                pretty_output = predict_request.send_csv(filepath=payload, host=host)

            logger.info(pretty_output)
            logger.info("\nPrediction Complete! ヽ(´▽`)/")
    except Exception as e:
        logger.error("{}".format(e))
        sys.exit(-1)

@click.command()
@click.option('--dir', help='Set the working directory for your mldock container.', required=True)
@click.option('--tag', help='docker tag', type=str, default='latest')
@click.option('--stage', default='dev', help='environment to stage.', required=True)
@click.pass_obj
def train(obj, dir, tag, stage):
    """
    Command to train ML model(s) locally
    """
    mldock_manager = MLDockConfigManager(
        filepath=os.path.join(dir, ".mldock.json")
    )
    # get mldock_module_path name
    mldock_config = mldock_manager.get_config()
    module_path = os.path.join(
        dir,
        mldock_config.get("mldock_module_dir", "src"),
    )
    image_name = mldock_config.get("image_name", None)
    platform = mldock_config.get("platform", None)
    container_dir = mldock_config.get("container_dir", None)
    helper_library_path = obj['helper_library_path']

    project_env_vars = mldock_config.get('environment', {})
    env_vars = collect_mldock_environment_variables(
        stage=stage,
        hyperparameters=mldock_config.get('hyperparameters', {}),
        **project_env_vars
    )
    try:
        if platform == "sagemaker":
            train_model(
                working_dir=dir,
                docker_tag=tag,
                image_name=image_name,
                entrypoint="container/executor.sh",
                cmd="train",
                env=env_vars
            )
        elif platform == "sagemakerv2":
            train_model(
                working_dir=dir,
                docker_tag=tag,
                image_name=image_name,
                entrypoint="src/container/executor.sh",
                cmd="train",
                env=env_vars
            )
        else:
            train_model(
                working_dir=dir,
                docker_tag=tag,
                image_name=image_name,
                entrypoint="src/container/executor.sh",
                cmd="train",
                env=env_vars
            )

        logger.info("Local training completed successfully!")
    except ValueError:
        logger.error("This is not a mldock directory: {}".format(dir))
        sys.exit(-1)
    except subprocess.CalledProcessError as e:
        logger.debug(e.output)
        raise
    except Exception as e:
        logger.error("{}".format(e))
        sys.exit(-1)

@click.command()
@click.option('--dir', help='Set the working directory for your mldock container.', required=True)
@click.option('--tag', help='docker tag', type=str, default='latest')
@click.option('--port', help='host url at which model is served', type=str, default='8080')
@click.option('--stage', default='dev', help='environment to stage.', required=True)
@click.pass_obj
def deploy(obj, dir, tag, port, stage):
    """
    Command to deploy ML model(s) locally
    """
    helper_library_path=obj['helper_library_path']
    mldock_manager = MLDockConfigManager(
        filepath=os.path.join(dir, ".mldock.json")
    )
    # get mldock_module_path name
    mldock_config = mldock_manager.get_config()
    module_path = os.path.join(
        dir,
        mldock_config.get("mldock_module_dir", "src"),
    )
    image_name = mldock_config.get("image_name", None)
    platform = mldock_config.get("platform", None)
    container_dir = mldock_config.get("container_dir", None)
    helper_library_path = obj['helper_library_path']

    project_env_vars = mldock_config.get('environment', {})
    env_vars = collect_mldock_environment_variables(
        stage=stage,
        hyperparameters=mldock_config.get('hyperparameters', {}),
        **project_env_vars
    )
    try:
        logger.info("Started local deployment at {}...\n".format(port))
        if platform == "sagemaker":
            deploy_model(
                working_dir=dir,
                docker_tag=tag,
                image_name=image_name,
                port=port,
                entrypoint="container/executor.sh",
                cmd="serve",
                env=env_vars
            )
        elif platform == "sagemakerv2":
            deploy_model(
                working_dir=dir,
                docker_tag=tag,
                image_name=image_name,
                port=port,
                entrypoint="src/container/executor.sh",
                cmd="serve",
                env=env_vars
            )
        else:
            deploy_model(
                working_dir=dir,
                docker_tag=tag,
                image_name=image_name,
                port=port,
                entrypoint="src/container/executor.sh",
                cmd="serve",
                env=env_vars
            )

    except ValueError:
        logger.error("This is not a mldock directory: {}".format(dir))
        sys.exit(-1)
    except subprocess.CalledProcessError as e:
        logger.debug(e.output)
        raise
    except Exception as e:
        logger.error("{}".format(e))
        sys.exit(-1)

local.add_command(build)
local.add_command(predict)
local.add_command(train)
local.add_command(deploy)



import os
import json
from pathlib import Path
import logging
import re

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(None)

def get_env_vars(regex, flat=True):
    """Get all environ vars matching contains='<PREFIX>'"""
    if flat:
        filtered_env_vars = []
    else:
        filtered_env_vars = {}

    regex_pattern = re.compile(regex)

    for key,value in os.environ.items():
        if regex_pattern.match(key):
            # get all keys matching the regex
            if flat:
                filtered_env_vars.append({'key': key, 'value': value})
            else:
                filtered_env_vars.append({key: value})
    return filtered_env_vars

class Environment:

    input_channel_regex = None
    model_channel_regex = None
    output_channel_regex = None
    output_channel_regex = None
    hyperparameters_file = None

    def __init__(self, base_dir, **kwargs):
        self.base_dir = base_dir
        self.hyperparameters_file = kwargs.get(
            'hyperparameters_file', 'hyperparameters.json'
        )

        self.input_channel_regex = kwargs.get(
            'input_channel_regex', 'INPUT_CHANNEL_.*'
        )
        self.model_channel_regex = kwargs.get(
            'model_channel_regex', 'MODEL_.*'
        )
        self.output_channel_regex = kwargs.get(
            'output_channel_regex', 'OUTPUT_DIR'
        )
        self.hyperparamters_regex = kwargs.get(
            'hyperparamters_regex', 'hyperparameters'
        )

        self._create_training_directories()
        self.setup_hyperparameters()

    def _create_training_directories(self):
        """Create the directory structure, if not exists
        """
        logger.debug("Creating a new training folder under {} .".format(self.base_dir))

        try:
            self.input_data_dir.mkdir(parents=True, exist_ok=True)
            self.model_dir.mkdir(parents=True, exist_ok=True)
            self.input_config_dir.mkdir(parents=True, exist_ok=True)
            self.output_data_dir.mkdir(parents=True, exist_ok=True)
            if not self.hyperparameters_filepath.exists():
                self._write_json({}, self.hyperparameters_filepath)
        except Exception as exception:
            logger.error(exception)

    @property
    def input_dir(self):
        return Path(self.base_dir, 'input')

    @property
    def input_data_dir(self):
        return Path(self.input_dir, 'data')

    @property
    def input_config_dir(self):
        return Path(self.input_dir, 'config')

    @property
    def hyperparameters_filepath(self):
        return Path(self.input_config_dir, self.hyperparameters_file)

    @property
    def model_dir(self):
        return Path(self.base_dir, 'model')
    
    @property
    def output_data_dir(self):
        return Path(self.base_dir, 'output')

    @property
    def hyperparameters(self):
        """Returns: Iterable of the input channels
        """
        return self._read_json(
            self.hyperparameters_filepath
        )
    
    @staticmethod
    def _read_json(path):  # type: (str) -> dict
        """Read a JSON file.
        Args:
            path (str): Path to the file.
        Returns:
            (dict[object, object]): A dictionary representation of the JSON file.
        """
        with open(path, "r") as f:
            return json.load(f)

    @staticmethod
    def _write_json(obj, path):  # type: (object, str) -> None
        """Write a serializeable object as a JSON file."""
        with open(path, "w") as f:
            json.dump(obj, f)
    
    def setup_hyperparameters(self):
        """Retrieves the env vars matching hyperparameters regex and updates config"""
        hyperparameters = self._read_json(
            self.hyperparameters_filepath
        )

        updated_hparams = get_env_vars(
            regex=self.hyperparamters_regex,
            flat=False
        )

        # updated_hparams = json.loads(updated_hparams)
        
        hyperparameters.update(
            updated_hparams
        )

        self._write_json(
            obj=hyperparameters,
            path=self.hyperparameters_filepath
        )

    def get_input_channel_iter(self):
        """Returns: Iterable of the input channels
        """
        return get_env_vars(regex=self.input_channel_regex)

    def get_output_channel_iter(self):
        """Returns: Iterable of the output channel
        """
        return get_env_vars(regex=self.output_channel_regex)

    def get_model_channel_iter(self):
        """Returns: Iterable of the model channel
        """
        return get_env_vars(regex=self.model_channel_regex)



import os
import sys
import json
import logging
import click
from pathlib import Path
from future.moves import subprocess

from mldock.config.config_manager import \
    ResourceConfigManager, MLDockConfigManager, PackageConfigManager, \
        HyperparameterConfigManager, InputDataConfigManager, WorkingDirectoryManager, \
            EnvironmentConfigManager
from mldock.api.local import \
    _copy_boilerplate_to_dst, _rename_file

from mldock.api.platform import sagemaker_init, generic_init

click.disable_unicode_literals_warning = True
logger=logging.getLogger('mldock')

@click.group()
def container():
    """
    Manage and create new container projects. Create new templates from mldock container projects.
    """
    pass

@click.command()
@click.option('--name', help='Container name, used to name directories,etc.', required=True, type=str)
@click.option('--dir', help='Relative name of MLDOCK project', required=True, type=str)
@click.option('--out', help='Destination of template should be stored once created.', required=True, type=str)
@click.pass_obj
def create_template(obj, name, dir, out):
    """
    Creates an MLDock ready container template

    remote => (giturl) => https://github.com/sdushantha/gitdir/blob/master/gitdir/gitdir.py
    """
    helper_library_path=obj['helper_library_path']

    try:
        if not Path(dir, '.mldock.json').exists():
            raise Exception("Path '{}' was not an mldock project. Confirm this directory is correct, otherwise create one.".format(dir))

        mldock_src_path = Path(dir, 'src')

        destination_path = Path(out, name)
        destination_path.mkdir(parents=False, exist_ok=True)

        destination_src_path = Path(destination_path, 'src')

        _copy_boilerplate_to_dst(mldock_src_path, destination_src_path)
    except Exception as exception:
        logger.error(exception)

@click.command()
@click.option('--dir', help='Set the working directory for your sagify container.', required=True)
@click.option('--testing_framework', default=None, help='(Optional) Pytest framework. This creates a few health-check tests')
@click.option('--service', default=None, help='(Optional) Docker Compose. This seeds a service config.')
@click.option('--no-prompt', is_flag=True, help='Do not prompt user, instead use the mldock config to initialize the container.')
@click.option('--container-only', is_flag=True, help='Only inject new container assets.')
@click.pass_obj
def init(obj, dir, testing_framework, service, no_prompt, container_only):
    """
    Command to initialize container configs required by sagemaker-training.
    """
    helper_library_path=obj['helper_library_path']
    try:
        logger.info("Initializing MLDock config")
        mldock_manager = MLDockConfigManager(
            filepath=os.path.join(dir, ".mldock.json")
        )
        if not no_prompt:
            mldock_manager.setup_config()

        package_manager = PackageConfigManager(
            filepath=os.path.join(dir, "requirements.txt")
        )
        # write to package manager
        package_manager.write_file()


        path_to_payload = Path(os.path.join(dir, "payload.json"))
        if not path_to_payload.exists():
            path_to_payload.write_text(json.dumps({"feature1": 10, "feature2":"groupA"}))

        # get sagify_module_path name
        mldock_config = mldock_manager.get_config()
        src_directory = os.path.join(
            dir,
            mldock_config.get("mldock_module_dir", "src")
        )
        mldock_module_path = os.path.join(
            src_directory,
            'container'
        )

        working_directory_manager = WorkingDirectoryManager(base_dir=dir)
        # sagemaker
        config_path = working_directory_manager.input_config_dir
        # get hyperparameters
        # set input data channels
        input_data_channels = InputDataConfigManager(
            config=mldock_config.get('data', [])
        )

        # set hyperparameters
        hyperparameters = HyperparameterConfigManager(
            config=mldock_config.get('hyperparameters', {})
        )

        # set environment variables
        environment = EnvironmentConfigManager(
            config=mldock_config.get('environment', {})
        )

        if not no_prompt:
            input_data_channels.ask_for_input_data_channels()
            mldock_manager.update_data_channels(data=input_data_channels.get_config())
            hyperparameters.ask_for_hyperparameters()
            mldock_manager.update_hyperparameters(hyperparameters=hyperparameters.get_config())
            environment.ask_for_env_vars()
            mldock_manager.update_env_vars(environment=environment.get_config())

        mldock_manager.write_file()
        mldock_manager.pretty_print()
        # Get platform specific files
        platform = mldock_config.get("platform", None)

        if platform == "sagemaker":
            # pass down any information like hyperparameters, etc
            sagemaker_init(
                dir=working_directory_manager.get_working_dir(),
                helper_library_path=helper_library_path,
                mldock_config=mldock_config,
                no_prompt=no_prompt,
                testing_framework=testing_framework,
                service=service,
                container_only=container_only
            )
        elif platform == "sagemakerv2":
            # pass down any information like hyperparameters, etc
            sagemaker_init(
                dir=working_directory_manager.get_working_dir(),
                helper_library_path=helper_library_path,
                mldock_config=mldock_config,
                no_prompt=no_prompt,
                testing_framework=testing_framework,
                service=service,
                container_only=container_only
            )
        elif platform == "generic":
            # pass down any information like hyperparameters, etc
            generic_init(
                dir=working_directory_manager.get_working_dir(),
                helper_library_path=helper_library_path,
                mldock_config=mldock_config,
                testing_framework=testing_framework,
                service=service,
                container_only=container_only
            )
        else:
            raise Exception("Platform not found. Please supply available platform template.")

        logger.info("\nlocal container volume is ready! ヽ(´▽`)/")
    except subprocess.CalledProcessError as e:
        logger.debug(e.output)
        raise
    except Exception as e:
        logger.error("{}".format(e))
        sys.exit(-1)

container.add_command(init)
container.add_command(create_template)



import os
from pathlib import Path
import logging

from mldock.platform_helpers.environment import Environment, get_env_vars

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(None)

class CustomEnvironment(Environment):
    """
        Extends the Environment class to give us
        more specific environment setup functionality
    """

    def setup_inputs(self):
        """ Iterates and downloads assets remoate -> input channels
        """
        for channel in self.get_input_channel_iter():
            logger.info("Setup Inputs: Must implement download scripts from remote")

    def cleanup_outputs(self):
        """ Iterates and uploads output channel -> remote
        """
        for channel in self.get_output_channel_iter():
            logger.info("Cleanup Outputs: Must implement download scripts to remote")

    def setup_model_artifacts(self):
        """ Iterates and downloads assets remoate -> model channel
        """
        for channel in self.get_model_channel_iter():
            logger.info("Setup Model: Must implement download scripts from remote")

    def cleanup_model_artifacts(self):
        """ Iterates and uploads from model channel -> remote
        """
        for channel in self.get_model_channel_iter():
            logger.info("Cleanup Model: Must implement download scripts to remote")

class TrainingContainer:
    """
        A set of tasks for setup and cleanup of container
    """
    def __init__(self):
        self.environment = CustomEnvironment(base_dir='/opt/ml')

    def startup(self):
        logger.info("\n\n --- Running Startup Script ---")
        logger.debug("Setting Up Training Container")
        environment = CustomEnvironment(base_dir='/opt/ml')
        if os.environ.get('stage', 'dev') == "prod":
            logger.debug("Env == Prod")
            self.environment.setup_inputs()
            self.environment.setup_model_artifacts()
        logger.info("--- Setup Complete --- \n\n")

    def cleanup(self):
        """clean up tasks executed on container task complete"""
        logger.info("\n\n --- Running Cleanup Script ---")
        logger.info("Cleaning Up Training Container")
        if os.environ.get('stage', 'dev') == "prod":
            logger.info("Env == Prod")
            self.environment.setup_inputs()
            self.environment.cleanup_model_artifacts()
        logger.info("--- Cleanup Complete --- \n\n")

class ServingContainer:
    """
        A set of tasks for setup and cleanup of container
    
        note:
            - Only supports a startup script. Cleanup is a bit fuzzy for serving.
    """
    def __init__(self):
        self.environment = CustomEnvironment(base_dir='/opt/ml')

    def startup(self):
        logger.info("\n\n --- Running Startup Script ---")
        logger.info("Setting Up Training Container")
        
        if os.environ.get('stage', 'dev') == "prod":
            logger.info("Env == Prod")
            self.environment.setup_model_artifacts()
        logger.info("\n\n --- Setup Complete --- \n\n")

    def cleanup(self):
        """clean up tasks executed on container task complete"""
        pass



import os
import json
import logging
from pathlib import Path
from future.moves import subprocess

from mldock.config.config_manager import \
    ResourceConfigManager, PackageConfigManager, \
        HyperparameterConfigManager, InputDataConfigManager
from mldock.api.local import \
    _copy_boilerplate_to_dst, _rename_file

logger=logging.getLogger('mldock')

def _write_json(obj, path):  # type: (object, str) -> None
    """Write a serializeable object as a JSON file."""
    with open(path, "w") as f:
        json.dump(obj, f)

def generic_init(dir, helper_library_path, mldock_config, testing_framework, service, container_only):

    logger.info("Initializing Assets - Generic Container")
    src_directory = os.path.join(
        dir,
        mldock_config.get("mldock_module_dir", "src")
    )
    service_directory = os.path.join(
        dir,
        'service'
    )

    logger.info("getting template")
    template_dir = mldock_config.get("template_dir", None)

    if template_dir is None:
        template_dir = os.path.join(
            helper_library_path,
            "templates",
            mldock_config['platform']
        )
        logger.debug("--- no template dir provided, using = (Default) {}".format(mldock_config['platform']))

    logger.debug("Template directory = {}".format(template_dir))


    mldock_module_path = os.path.join(
        src_directory,
        'container'
    )

    test_path = os.path.join(mldock_module_path, 'local_test', 'test_dir')

    logger.info("Setting up workspace")   
    logger.debug("Adding container")
    _copy_boilerplate_to_dst(
        os.path.join(template_dir,'src/container'),
        mldock_module_path,
        remove_first=True
    )
    if container_only == False:
        logger.debug("Adding src scripts")
        _copy_boilerplate_to_dst(os.path.join(template_dir,'src/'), src_directory)

    if testing_framework == 'pytest':
        logger.info("Adding pytest container tests")
        _copy_boilerplate_to_dst(os.path.join(template_dir,'tests/'), 'tests/')
        logger.debug("renaming test file")

        _rename_file(
            base_path='tests/container_health',
            current_filename='_template_test_container.py',
            new_filename='test_{ASSET_DIR}.py'.format(ASSET_DIR=dir)
        )
    if service is not None:
        logger.info("Adding {} service".format(service))
        _copy_boilerplate_to_dst(os.path.join(template_dir,'service',service), service_directory)

def _make_sagemaker_input_data_configs(data_channels: list):
    """restructures a list of data names as sagemaker input data configs"""
    return {name: {} for name in data_channels}

def _extract_data_channels_from_mldock(mldock_data: list):

    data_channels = [
        data_config['channel'] for data_config in mldock_data
    ]

    return _make_sagemaker_input_data_configs(data_channels=data_channels)

def sagemaker_init(dir, helper_library_path, mldock_config, no_prompt, testing_framework, service, container_only):

    logger.info("Initializing Assets - Sagemaker Container")
    src_directory = os.path.join(
        dir,
        mldock_config.get("mldock_module_dir", "src")
    )

    service_directory = os.path.join(
        dir,
        'service'
    )

    logger.info("getting template")
    template_dir = mldock_config.get("template_dir", None)

    if template_dir is None:
        template_dir = os.path.join(
            helper_library_path,
            "templates",
            mldock_config['platform']
        )
        logger.debug("--- no template dir provided, using = (Default) {}".format(mldock_config['platform']))

    logger.debug("Template directory = {}".format(template_dir))

    mldock_module_path = os.path.join(
        src_directory,
        'container'
    )

    test_path = os.path.join(mldock_module_path, 'local_test', 'test_dir')

    logger.info("Setting up workspace")   
    logger.debug("Adding container")
    _copy_boilerplate_to_dst(
        os.path.join(template_dir,'src/container'),
        mldock_module_path,
        remove_first=True
    )
    if container_only == False:
        logger.debug("Adding src scripts")
        _copy_boilerplate_to_dst(os.path.join(template_dir,'src/'), src_directory)

    if testing_framework == 'pytest':
        logger.info("Adding pytest container tests")
        _copy_boilerplate_to_dst(os.path.join(template_dir,'tests/'), 'tests/')
        logger.debug("renaming test file")

        _rename_file(
            base_path='tests/container_health',
            current_filename='_template_test_container.py',
            new_filename='test_{ASSET_DIR}.py'.format(ASSET_DIR=dir)
        )

    if service is not None:
        logger.info("Adding {} service".format(service))
        _copy_boilerplate_to_dst(os.path.join(template_dir, 'service', service), service_directory)

    logger.info("Adding dummy sagemaker goodies")
    _write_json({}, Path(dir,'config/hyperparameters.json'))

    # get list of dataset names
    mldock_data = mldock_config.get("data", None)
    input_config_config = _extract_data_channels_from_mldock(mldock_data)
    _write_json(input_config_config, Path(dir,'config/inputdataconfig.json'))
    _write_json({"current_host": "dummy_host", "hosts": ["dummy_host", "second_dummy_host"]}, Path(dir,'config/resourceconfig.json'))



import setuptools
from mldock.__version__ import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mldock",
    version=__version__,
    author="SheldonGrant",
    author_email="sheldz.shakes.williams@gmail.com",
    description="Global Machine learning helpers for docker based development. Build, train and deploy on cloud with docker",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SheldonGrant/locative-ml-global-helpers",
    packages=setuptools.find_packages(where='.'),
    package_data={
        'mldock': [
            'templates/sagemaker/tests/container_health/*.py',
            'templates/sagemaker/src/',
            'templates/sagemaker/src/*.py',
            'templates/sagemaker/src/utils',
            'templates/sagemaker/src/utils/*py',
            'templates/sagemaker/src/container',
            'templates/sagemaker/src/container/*.sh',
            'templates/sagemaker/src/container/Dockerfile',
            'templates/sagemaker/src/container/ __init__.py',
            'templates/sagemaker/src/container/training/__init__.py',
            'templates/sagemaker/src/container/training/train',
            'templates/sagemaker/src/container/prediction/*.py',
            'templates/sagemaker/src/container/prediction/serve',
            'templates/sagemaker/src/container/prediction/nginx.conf',
            'templates/sagemaker/service/compose/docker-compose.yml',
            'templates/sagemakerv2/src/',
            'templates/sagemakerv2/src/*.py',
            'templates/sagemakerv2/src/utils',
            'templates/sagemakerv2/src/utils/*py',
            'templates/sagemakerv2/src/container',
            'templates/sagemakerv2/src/container/*.sh',
            'templates/sagemakerv2/src/container/Dockerfile',
            'templates/sagemakerv2/src/container/ __init__.py',
            'templates/sagemakerv2/src/container/training/__init__.py',
            'templates/sagemakerv2/src/container/training/train',
            'templates/sagemakerv2/src/container/prediction/*.py',
            'templates/sagemakerv2/src/container/prediction/serve',
            'templates/sagemakerv2/src/container/prediction/nginx.conf',
            'templates/sagemakerv2/service/compose/docker-compose.yml',
            'templates/generic/tests/container_health/*.py',
            'templates/generic/service/compose/docker-compose.yml',
            'templates/generic/src/',
            'templates/generic/src/environment.py',
            'templates/generic/src/prediction.py',
            'templates/generic/src/trainer.py',
            'templates/generic/src/container',
            'templates/generic/src/container/*.sh',
            'templates/generic/src/container/Dockerfile',
            'templates/generic/src/container/ __init__.py',
            'templates/generic/src/container/training/',
            'templates/generic/src/container/training/train.py',
            'templates/generic/src/container/prediction/',
            'templates/generic/src/container/prediction/wsgi.py',
            'templates/generic/src/container/prediction/predictor.py',
            'templates/generic/src/container/prediction/nginx.conf'
        ]
    },
    setup_requires=['setuptools>=39.1.0', 'future'],
    extras_require={
        'ai-platform': ['google-cloud-storage', 'google-api-python-client'],
        'cli': ['click', 'docker', 'future', 'requests'],
        'sagemaker': ['boto3', 'sagemaker-training']
    },
    entry_points="""
        [console_scripts]
        mldock=mldock.__main__:cli
    """,
    keywords=["docker", "machine learning", "ml", "ml services", "MLaaS"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6"
    ],
    python_requires='>=3.6',
)



# This is the file that implements a flask server to do inferences. It's the file that you will
#  modify to implement the scoring for your own algorithm.

from __future__ import print_function, absolute_import

# Do not remove the following line
import sys;sys.path.insert(1, ".")  # Do not remove this

import json

import flask

# from . import predict
import src.prediction as predict_function


app = flask.Flask(__name__)


@app.route('/ping', methods=['GET'])
def ping():
    """Determine if the container is working and healthy"""
    return flask.Response(response='\n', status=200, mimetype='application/json')


@app.route('/invocations', methods=['POST'])
def transformation():
    """Do an inference on a single batch of data. In this sample server, we take data as JSON"""
    if flask.request.content_type == 'application/json':
        data = flask.request.get_json()
    else:
        return flask.Response(
            response=json.dumps({'message': 'This predictor only supports JSON data'}),
            status=415,
            mimetype='application/json'
        )

    result = predict_function.predict(data)

    return flask.Response(response=json.dumps(result), status=200, mimetype='application/json')



"""
    PLATFORM HELPERS

    utils used to interact with specific platform from a docker perspective.
    E.g. cloud storage tools to upload and download objects.
"""
import os
import glob
from google.cloud import storage
from pathlib import Path
import tarfile
import zipfile
from typing import Iterator
import logging
from urllib.parse import urlparse

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(None)
logger.setLevel(logging.INFO)

def _mkdir(dir_path: str):
    """make directory structure

    Args:
        dir_path (str): directory path to build
    """
    p = Path(dir_path)
    p.mkdir(parents=True, exist_ok=True)

def _iter_nested_dir(root_dir: str) -> Iterator[str]:
    """Iterate through nested folders.

    Args:
        root_dir (str): name of the directory to start form

    Yields:
        Iterator[str]: path iterator
    """
    rootdir = Path(root_dir)
    for path in rootdir.glob('**/*'):
        yield path

def _delete_file(file_path):
    """Delete a file

    Args:
        file_path (str): path to file

    Raises:
        TypeError: Expected a file_path of type file. Cannot be directory or other.
    """
    path = Path(file_path)
    if not path.is_file():
        raise TypeError("Expected a file_path of type file. Only deletes files")

    os.remove(path)

def _check_if_cloud_scheme(url: str, scheme: str = 's3') -> bool:
    parsed_url = urlparse(url)
    return parsed_url.scheme == scheme

def parse_url(url: str, scheme: str = 's3'):
    """Returns an (s3 bucket, key name/prefix) tuple from a url with an s3
    scheme.
    Args:
        url (str):
    Returns:
        tuple: A tuple containing:
            - str: S3 bucket name
            - str: S3 key
    """
    parsed_url = urlparse(url)
    if parsed_url.scheme != scheme:
        raise ValueError("Expecting '{}' scheme, got: {} in {}.".format(scheme, parsed_url.scheme, url))
    return parsed_url.netloc, parsed_url.path.lstrip("/")

def zip_folder(dir_path, output_file, rm_original=True):
    """zip in directory and optionally throw away unzipped"""

    with zipfile.ZipFile(output_file, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for file in _iter_nested_dir(dir_path):
            if file.name != Path(output_file).name and file.is_file():
                zipf.write(os.path.join(file), arcname=file.relative_to(dir_path))
                if rm_original:
                    _delete_file(file)

def zip_folder_as_tarfile(dir_path, output_file, rm_original=True):
    """zip folder as tarfile and optionally throw away original files"""
    with tarfile.open(output_file, "w:gz") as tar:
        for file in _iter_nested_dir(dir_path):
            if file.name != Path(output_file).name and file.is_file():
                tar.add(file, arcname=file.relative_to(dir_path))
                if rm_original:
                    _delete_file(file)

def unzip_file(filename, output_dir, rm_zipped=True):
    """unzip in directory and optionally throw away zipped"""
    with zipfile.ZipFile(filename, 'r', zipfile.ZIP_DEFLATED) as zipf:
        logger.info("Unzipping {} => {}".format(filename, output_dir))
        zipf.extractall(output_dir)
        if rm_zipped:
            logger.info("Removing {}".format(filename))
            _delete_file(filename)

def unzip_file_from_tarfile(filename, output_dir, rm_zipped=True):
    """untar in directory and optionally throw away zipped"""
    with tarfile.open(filename, "r:gz") as tar:
        logger.info("Zipping {} => {}".format(filename, output_dir))
        tar.extractall(output_dir)
        if rm_zipped:
            logger.info("Removing {}".format(filename))
            _delete_file(filename)

def download_folder(
    bucket_name: str,
    prefix: str,
    target: str
):
    """download folder from google cloud storage, ignoring folders.

    Args:
        bucket_name (str): [description]
        prefix (str): [description]
        target (str): [description]
    """
    logger.info("Starting Folder download from cloud storage")
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)
    blobs = bucket.list_blobs(prefix=prefix)  # Get list of files
    for blob in blobs:
        filename = blob.name.rsplit('/',1)[1]#.replace('/', '_')

        if len(filename) > 0:
            
            fullpath = os.path.join(bucket_name, prefix, filename)
            dst_filepath = os.path.join(target, os.path.basename(prefix))
            file_destination = os.path.join(dst_filepath, filename)

            logger.info("Download {} to local at {}".format(filename, file_destination))
            # make directory
            _mkdir(dst_filepath)
            # download
            blob.download_to_filename(file_destination)

def upload_folder(local_path: str, bucket_name: str, prefix: str):
    """Upload folder and contents to cloud storage

    Args:
        local_path (str): path to local directory
        bucket_name (str): bucket name to upload files to
        prefix (str): cloud filepath to write files to
    """
    logger.info("Starting Folder Upload to cloud storage")
    # get bucket
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)
    assert os.path.isdir(local_path)
    for path in _iter_nested_dir(local_path):
        filepath = path.relative_to(local_path)
        storage_destination = os.path.join(prefix, filepath)
        if path.is_file():
            logger.info("Upload {} to cloud at {}".format(path, storage_destination))
            blob = bucket.blob(storage_destination)
            blob.upload_from_filename(path)

def download_input_assets(storage_dir_path: str, local_path: str, scheme: str = 'gs'):
    """download input asset folder from path, given that path is cloud storage path.

    Args:
        storage_dir_path (str): path to data in cloud storage
        local_path (str): where data should be stored (this comes from the environment).
        scheme (str, optional): scheme for cloud storage url. Defaults to 'gs'.
    """
    is_cloud_storage = _check_if_cloud_scheme(url=storage_dir_path, scheme=scheme)

    if is_cloud_storage:
        # download data if cloud storage path
        bucket, prefix = parse_url(url=storage_dir_path, scheme=scheme)
        download_folder(bucket_name=bucket, prefix=prefix, target=local_path)
    else:
        Exception("No Cloud storage url was found. Must have gs:// schema")

def package_and_upload_model_dir(local_path: str, storage_dir_path: str, scheme: str = 'gs'):
    """
        Packages model/ dir as .tar.gz and uploads to cloud storage, given that storage_dir_path
        is cloud storage url path.

        Args:
            local_path (str): path to model directory
            storage_dir_path (str): cloud storage path destination
            scheme (str, optional): cloud storage url scheme. Defaults to 'gs'.
    """
    is_cloud_storage = _check_if_cloud_scheme(url=storage_dir_path, scheme=scheme)
    if is_cloud_storage:
        zip_folder_as_tarfile(
            dir_path=local_path,
            output_file=os.path.join(local_path, 'model.tar.gz'),
            rm_original=True
        )
        bucket, prefix = parse_url(url=storage_dir_path, scheme=scheme)
        upload_folder(local_path=local_path, bucket_name=bucket, prefix=prefix)
    else:
        Exception("No Cloud storage url was found. Must have gs:// schema")

def package_and_upload_output_data_dir(local_path: str, storage_dir_path: str, scheme: str = 'gs'):
    """
        Packages output/ dir as .tar.gz and uploads to cloud storage, given that storage_dir_path
        is cloud storage url path.

        Args:
            local_path (str): path to model directory
            storage_dir_path (str): cloud storage path destination
            scheme (str, optional): cloud storage url scheme. Defaults to 'gs'.
    """
    is_cloud_storage = _check_if_cloud_scheme(url=storage_dir_path, scheme=scheme)

    if is_cloud_storage:
        bucket, prefix = parse_url(url=storage_dir_path, scheme=scheme)
        upload_folder(local_path=local_path, bucket_name=bucket, prefix=prefix)
    else:
        Exception("No Cloud storage url was found. Must have gs:// schema")
